﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SharpGL.SceneGraph;
using System.IO;
using System.Text.RegularExpressions;
using SharpGL;

namespace LindenmayerTool
{
    // LSystem
    // Holds all L-System configuration settings,
    // manages L-System string replacement and relays LTurtle operations.
    public class LSystem
    {
        // Random number generator and optional seed value
        public Random RndGen = new Random();
        public Nullable<int> RndGenSeed = null;

        // L-System configuration
        public String Seed;                                             // Initial axiom
        public List<ProductionRule> Rules = new List<ProductionRule>(); // List of production rules
        public List<String> UserDefTokens = new List<String>();         // List of found user-defined tokens (for debugging)
        public List<Vertex> ColourPalette = new List<Vertex>(); // List of colours available
        public float StepSize;                                          // Length of each forward line segment
        public float RotationAngle;                                     // Angle to rotate on rotation commands
        public float SegmentWidthInitial;                               // Initial segment width
        public float SegmentWidthDecrement;                             // Proportion of width to decrement
        public int Iterations;                                          // Number of replacement iterations

        public String FinalString = "";                                 // A stored copy of the final L-System string
        public List<DrawObject> Geometry = new List<DrawObject>();      // A stored copy of all interpreted lines/geometry
        
        protected LTurtle Turtle = new LTurtle();                                 // The active LTurtle
        protected Stack<LTurtle> TurtleStack = new Stack<LTurtle>();              // Stack of stored LTurtles

        public Nullable<float> RingAdjustment = null;                   //
        public Boolean CurveAdjustOnForward = true;                    // Stores rotational data and creates a curve on "forward" command instead of an instant turn.
        //TODO:
        // proper implementation will require a list to keep track of new turn data received.
        //public int ForwardRotationShare = 1;                            // Amount of "forwards" commands that is required to neutralize turn data

        // The (program-wide) active LSystem and getter/setter
        private static LSystem ActiveSystem = null;
        public static LSystem I()
        {
            if (ActiveSystem == null) ActiveSystem = new LSystem_Musical(); //LSystem();
            return ActiveSystem;
        }
        public static void SetActive(LSystem sys)
        {
            ActiveSystem = sys;
        }

        public LSystem()
        {
            Seed = "S";
            StepSize = 5.0f;
            RotationAngle = 90.0f;
            SegmentWidthInitial = 1.0f;
            SegmentWidthDecrement = 0.1f;
            Iterations = 10;
            ColourPalette.Add(new Vertex(1,1,1));
            ColourPalette.Add(new Vertex(1,0,0));
            ColourPalette.Add(new Vertex(0,1,0));
            ColourPalette.Add(new Vertex(0,0,1));
        }

        // Returns list of L-System tokens (characters, or characters with parametric expressions)
        protected MatchCollection GetTokenSet(String str)
        {
            Regex ParametricMatcher = new Regex(@"[^\(\)](\([^\(\)]+\))?"); // x || x(n)
            return ParametricMatcher.Matches(str);
        }

        // Returns the next iteration of the L-System
        public String GetNextItteration(String current)
        {
            StringBuilder sb = new StringBuilder();
            MatchCollection tokens = GetTokenSet(current);
            foreach (Match token in tokens)
            {
                var c = token.Value;
                // Get replacement rule
                var rule = Rules.Where(t => t.isMatch(c)).FirstOrDefault();
                if (rule == null)
                    // If no rule, then replace with itself
                    sb.Append(c);
                else
                    // Otherwise, use the rule to determine the replacement
                    sb.Append(rule.getReplacement(c.ToString()));
            }
            return sb.ToString();
        }

        // Gets the system specified n'th iteration of the system from the axiom.
        // (Or uses the stored final iteration if it exists)
        public String GetFinalItteration(bool useCache = true)
        {
            if (String.IsNullOrEmpty(FinalString) || !useCache )
            {
                String cur = Seed;
                for (int i = 0; i < Iterations; i++)
                    cur = GetNextItteration(cur);
                FinalString = cur;
            }
            return FinalString;
        }

        // For debugging; Has a turtle process steps n to m
        public void DoSteps(String str, int start, int end)
        {
            string substr = str.Substring(start, end - start);
            Do(substr);
        }

        // Does all processing of a system
        // Resets RNG, recomputes final iteration of L-System string, and sends all commands to turtle
        public void DoAll()
        {
            if (RndGenSeed != null)
                RndGen = new Random(RndGenSeed.Value);
            else
                RndGen = new Random();
            Do(this.GetFinalItteration(false));
        }

        // Sends commands from a L-System string to turtle
        public virtual void Do(String str)
        {
            // Clear stored lines/geometry
            Geometry.Clear();

            // Initialize new turtle
            Turtle = new LTurtle();
            Turtle.SegmentDiameter = SegmentWidthInitial;

            // Get a list of tokens, and process.
            var tokens = GetTokenSet(str);
            for (int i = 0; i < tokens.Count; i++)
            {
                var match = tokens[i];      // Current token (as a regex match)
                var token = match.Value;    // Current token expression
                var param = "";             // Current token parametric expression (if exists)
                if (match.Captures.Count > 1)
                    param = (match.Captures[1].Value).Substring(1, match.Captures[1].Value.Length - 2);
                var c = token[0];           // Current token variable/command
                // If the token is a command, send it to turtle
                if (Enum.IsDefined(typeof(LTurtle.Operation), (byte)c))
                {
                    switch ((LTurtle.Operation)c)
                    {
                        case LTurtle.Operation.BranchStart:
                            TurtleStack.Push(Turtle);
                            Turtle = new LTurtle(Turtle);
                            break;
                        case LTurtle.Operation.BranchEnd:
                            if (TurtleStack.Count > 0)
                                Turtle = TurtleStack.Pop();
                            break;
                        case LTurtle.Operation.CutBranch:
                                // Proceed to next branch end (or \0)
                                // Do no intermediate operations, do not pass go, do not collect $200.
                                for(int seek_i = i+1; i < str.Length; i++)
                                    if (str[seek_i].Equals(LTurtle.Operation.BranchEnd))
                                    {
                                        i = seek_i;
                                        break;
                                    }
                            break;
                        default:
                            Turtle.Do((LTurtle.Operation)c);
                            break;
                    }
                }
                // Otherwise, it's from leftover structural rules. Ignore.
                else if (!LSystem.I().UserDefTokens.Contains(c.ToString()))
                {
                    System.Console.WriteLine("Found invalid turtle command: '" + c + "'.");
                    LSystem.I().UserDefTokens.Add(c.ToString());
                }
            }
        }

        // Draw each piece of stored geometry (line)
        public void Render(OpenGL Renderer)
        {
            foreach (var dObj in Geometry)
                dObj.Render(Renderer);
        }

        // Returns an LSystem configured from a file
        public static LSystem LoadFromFile(String filename)
        {
            var sys = new LSystem_Musical();
            var file = File.ReadAllLines(filename);

            String[] RuleSideDelimeter = new String[] { "-->" };    // Separator of LHS and RHS
            Regex Regex_Seed   = new Regex(@"^w\s*:");              // Regex to match axiom tags
            Regex Regex_Rule   = new Regex(@"^p[0-9]*\s*:");        // Regex to match rule tags
            // Regex to match a LHS/RHS separator that includes a weight for stochastic systems
            Regex Regex_Weight = new Regex(@"\(\s*([0-9\.]*)\s*\)\s*"+RuleSideDelimeter.Single());

            bool ColoursDefined = false;

            foreach (var rawLine in file)
            {
                var s = rawLine.Trim();

                // TODO: (advanced definitions for prefab geometry and parametric systems)
                if (s.StartsWith("#define")) continue;
                if (s.StartsWith("#ignore")) continue;
                if (s.StartsWith("#include")) continue;
                
                if ( s.StartsWith("#") || s.Equals("") ) continue;   // comments

                if (s.StartsWith("Iterations"))
                {
                    try { Int32.TryParse((s.Split(":".ToCharArray())).Skip(1).FirstOrDefault().Trim(), out sys.Iterations); }
                    catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                    continue;
                }
                if (s.StartsWith("StepSize"))
                {
                    try { float.TryParse((s.Split(":".ToCharArray())).Skip(1).FirstOrDefault().Trim(), out sys.StepSize); }
                    catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                    continue;
                }
                if (s.StartsWith("AngleIncrement"))
                {
                    try { float.TryParse((s.Split(":".ToCharArray())).Skip(1).FirstOrDefault().Trim(), out sys.RotationAngle); }
                    catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                    continue;
                }
                if (s.StartsWith("WidthStart"))
                {
                    try { float.TryParse((s.Split(":".ToCharArray())).Skip(1).FirstOrDefault().Trim(), out sys.SegmentWidthInitial); }
                    catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                    continue;
                }
                if (s.StartsWith("WidthDecrement"))
                {
                    try { float.TryParse((s.Split(":".ToCharArray())).Skip(1).FirstOrDefault().Trim(), out sys.SegmentWidthDecrement); }
                    catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                    continue;
                }
                if (s.StartsWith("Colour"))
                {
                    if(!ColoursDefined)
                        sys.ColourPalette.Clear();
                    ColoursDefined = true;
                    var ColourStr = s.Split(":".ToCharArray()).Skip(1).FirstOrDefault().Trim();
                    var Colours = ColourStr.Split(",".ToCharArray());
                    foreach (var c in Colours)
                    {
                        // Parse RGB hex string to float vertex.
                        uint iColour = 0;
                        try { iColour = Convert.ToUInt32(c.Trim(), 16); }
                        catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                        var vColour = new Vertex();
                        vColour.X = (float)((iColour & 0xFF0000) >> 16) / 255;
                        vColour.Y = (float)((iColour & 0x00FF00) >> 8 ) / 255;
                        vColour.Z = (float)((iColour & 0x0000FF)      ) / 255;
                        sys.ColourPalette.Add(vColour);
                    }
                    continue;
                }

                // Seed?
                if(Regex_Seed.Match(s).Success)
                    sys.Seed = s.Split(":".ToCharArray()).Skip(1).FirstOrDefault().Trim();

                // Rule?
                if (Regex_Rule.Match(s).Success)
                {
                    var tag = s.Split(":".ToCharArray()).FirstOrDefault();                      // Tag
                    var s2 = s.Substring(tag.Length).Trim();
                    var ruleArr = s2.Split(":".ToCharArray());
                    var ruleCondition = (ruleArr.Length > 1) ? ruleArr.First().Trim() : "";     // Parametric conditions
                    var ruleStr = ruleArr.Last().Trim();
                    Match weightMatch = Regex_Weight.Match(ruleStr);
                    float ruleWeight = 1;
                    if (weightMatch.Success)
                    {
                        try { float.TryParse(weightMatch.Groups[1].Value.Trim(), out ruleWeight); }
                        catch (Exception e) { System.Console.WriteLine("Error: " + e.ToString()); }
                        ruleStr = ruleStr.Replace(weightMatch.Value, RuleSideDelimeter.Single());
                    }
                    var ruleStrArr = ruleStr.Split(RuleSideDelimeter, StringSplitOptions.None);
                    var rule_LHS = ruleStrArr.First().Trim();
                    var rule_RHS = ruleStrArr.Last().Trim();

                    // Does a rule with the same LHS exist? Stochastic.
                    ProductionRule rule = sys.Rules.Where(r => r.LHS == rule_LHS).FirstOrDefault();
                    if (rule == null)
                    {
                        rule = new ProductionRule(tag);
                        sys.Rules.Add(rule);
                        rule.LHS = rule_LHS;
                    }
                    rule.RHS.Add(new Tuple<float, string>( ruleWeight, rule_RHS ));
                }

            }
            return sys;
        }

        public void ApplyWrap(WrapOptions.Axis axis, WrapOptions.OverlapScales scale, float overlap, bool isProjection)
        {
            float ymin = float.MaxValue;
            float ymax = float.MinValue;
            foreach(DrawObject dObj in Geometry)
            {
                var d = (DrawObject.DO_Line)dObj;
                if(d.P1.Y < ymin) ymin = d.P1.Y;
                if(d.P2.Y < ymin) ymin = d.P2.Y;
                if(d.P1.Y > ymax) ymax = d.P1.Y;
                if(d.P2.Y > ymax) ymax = d.P2.Y;
            }
            float linearlength = ymax - ymin;
            
            //double sAng = (2 * Math.PI) / linearlength;
            double sAng = 360.0 / linearlength;

            RingAdjustment = (float)sAng * (overlap/100);
            DoAll();
        }
        public void UnApplyWrap()
        {
            RingAdjustment = null;
            DoAll();
        }

    }
}
