﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SharpGL;
using System.ComponentModel;

namespace LindenmayerTool
{
    // Extends the LSystem class with methods and members to interpret audio events
    public class LSystem_Musical : LSystem
    {

        // A stored copy of all audio events
        public List<SoundObject> Score = new List<SoundObject>();

        // Metric used to determine when an audio event starts
        public TimeMetric TimingMetric = TimeMetric.TotalStepLength;
        public enum TimeMetric
        {
            [Description("Total Step Length")]
            TotalStepLength,
            [Description("3D (XYZ) Origin Distance")]
            OriginDistance3D,
            [Description("2D (XY ) Origin Distance")]
            OriginDistance2D,
            [Description("X Axis Position")]
            Coord_X,
            [Description("Y Axis Position")]
            Coord_Y,
            [Description("Z Axis Position")]
            Coord_Z
        }

        // The length (in millis) that each step holds before advancing
        public float StepTimeScale = 100;

        // Is the sound event created before or after the turtle updates
        public NoteStepCreationMetric NoteStepCreateMetric = NoteStepCreationMetric.StartStep;
        public enum NoteStepCreationMetric
        {
            [Description("Start of Step")]
            StartStep,
            [Description("End of Step")]
            EndStep
        }

        public enum IncrementOrAxis
        {
            [Description("Adjusted by Operations")]
            Increment,
            [Description("X Axis Position")]
            Coord_X,
            [Description("Y Axis Position")]
            Coord_Y,
            [Description("Z Axis Position")]
            Coord_Z
        }

        // Default value, amount adjusted, adjustment scale, and value source of
        // Pitch, Duration, and Volume
        public float PitchDefault = 60;
        public float PitchAdjustment = 1;
        public AdjustmentType PitchAdjustmentType = AdjustmentType.Additive;
        public IncrementOrAxis PitchDeterminationType = IncrementOrAxis.Increment;
        public float DurationDefault = 250;
        public float DurationAdjustment = 0.10f;
        public AdjustmentType DurationAdjustmentType = AdjustmentType.Multiplicative;
        public IncrementOrAxis DurationDeterminationType = IncrementOrAxis.Increment;
        public float VolumeDefault = 10000;
        public float VolumeAdjustment = 0.10f;
        public AdjustmentType VolumeAdjustmentType = AdjustmentType.Multiplicative;
        public IncrementOrAxis VolumeDeterminationType = IncrementOrAxis.Increment;

        public enum AdjustmentType
        {
            [Description("Linear (Additive)")]
            Additive,
            [Description("Logarithmic (Multiplicative)")]
            Multiplicative
        }

        // Default value, amount adjusted, adjustment scale, and value source of
        // custom parameters
        public float ParamC1Default = 0;
        public float ParamC2Default = 0;
        public float ParamC3Default = 0;
        public float ParamC1Adjustment = 0.10f;
        public float ParamC2Adjustment = 0.10f;
        public float ParamC3Adjustment = 0.10f;
        public AdjustmentType ParamC1AdjustmentType = AdjustmentType.Multiplicative;
        public AdjustmentType ParamC2AdjustmentType = AdjustmentType.Multiplicative;
        public AdjustmentType ParamC3AdjustmentType = AdjustmentType.Multiplicative;
        public IncrementOrAxis ParamC1DeterminationType = IncrementOrAxis.Increment;
        public IncrementOrAxis ParamC2DeterminationType = IncrementOrAxis.Increment;
        public IncrementOrAxis ParamC3DeterminationType = IncrementOrAxis.Increment;

        // Use absolute values for axis sources, or allow negatives?
        public Boolean isAxisValuesAbsolute = true;


        // Override the default Do method to setup and track audio events
        // Creates a traversable tree of audio events
        public override void Do(String str)
        {

            // Clear stored lines/geometry and score
            Geometry.Clear();
            Score.Clear();

            var ls = (LSystem_Musical)(LSystem.I());

            // Initialize new turtle
            var CurrentTurtle = new LTurtle_Musical();
            Turtle = CurrentTurtle; 
            // Graphic Init
            CurrentTurtle.SegmentDiameter = SegmentWidthInitial;
            // Sound Init
            CurrentTurtle.Sound.Instrument = 0;
            CurrentTurtle.Sound.Pitch = PitchDefault;
            CurrentTurtle.Sound.Duration = DurationDefault;
            CurrentTurtle.Sound.Volume = VolumeDefault;
            CurrentTurtle.Sound.ParamC1 = ParamC1Default;
            CurrentTurtle.Sound.ParamC2 = ParamC2Default;
            CurrentTurtle.Sound.ParamC3 = ParamC3Default;

            // Get a list of tokens, and process.
            var tokens = GetTokenSet(str);
            for (int i = 0; i < tokens.Count; i++)
            {
                var match = tokens[i];      // Current token (as a regex match)
                var token = match.Value;    // Current token expression
                var param = "";             // Current token parametric expression (if exists)
                if (match.Captures.Count > 1)
                    param = (match.Captures[1].Value).Substring(1, match.Captures[1].Value.Length - 2);
                var c = token[0];           // Current token variable/command

                // If the token is a command, send it to turtle
                if (Enum.IsDefined(typeof(LTurtle.Operation), (byte)c))
                {
                    switch ((LTurtle.Operation)c)
                    {
                        case LTurtle.Operation.BranchStart:
                            {
                                TurtleStack.Push(CurrentTurtle);
                                CurrentTurtle = new LTurtle_Musical(CurrentTurtle);
                                break;
                            }
                        case LTurtle.Operation.BranchEnd:
                            {
                                if (TurtleStack.Count > 0)
                                    CurrentTurtle = (LTurtle_Musical)TurtleStack.Pop();
                                break;
                            }
                        case LTurtle.Operation.CutBranch:
                            {
                                // Proceed to next branch end (or \0)
                                // Do no intermediate operations, do not pass go, do not collect $200.
                                for (int seek_i = i + 1; i < str.Length; i++)
                                    if (str[seek_i].Equals(LTurtle.Operation.BranchEnd))
                                    {
                                        i = seek_i;
                                        break;
                                    }
                                break;
                            }
                        case LTurtle.Operation.MoveForwardAndDraw:
                            {
                                // MoveAndDraw --> New sound event --> new node in sound event tree

                                if (ls.NoteStepCreateMetric == NoteStepCreationMetric.EndStep)
                                {
                                    CurrentTurtle.Do((LTurtle.Operation)c);
                                    CurrentTurtle.DoSound((LTurtle.Operation)c);
                                }
                                
                                var NewTurtle = new LTurtle_Musical(CurrentTurtle);
                                CurrentTurtle.ChildBranches.Add(NewTurtle);
                                CurrentTurtle = NewTurtle;

                                if (ls.NoteStepCreateMetric == NoteStepCreationMetric.StartStep)
                                {
                                    CurrentTurtle.Do((LTurtle.Operation)c);
                                    CurrentTurtle.DoSound((LTurtle.Operation)c);
                                }
                                break;
                            }
                        default:
                            {
                                CurrentTurtle.Do(      (LTurtle.Operation)c);
                                CurrentTurtle.DoSound( (LTurtle.Operation)c);
                                break;
                            }
                    }
                }

                // Otherwise, it's from leftover structural rules. Ignore.
                else if (!LSystem.I().UserDefTokens.Contains(c.ToString()))
                {
                    System.Console.WriteLine("Found invalid turtle command: '" + c + "'.");
                    LSystem.I().UserDefTokens.Add(c.ToString());
                }

            }

        }//DO

        // Allow passing a timestamp to the render methods to let them know which
        // drawobjects to highlight
        public void Render(OpenGL Renderer, long curTime)
        {
            foreach (var dObj in Geometry)
                dObj.Render(Renderer, (int)curTime);
        }


    }
}
