﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SharpGL.SceneGraph;

namespace LindenmayerTool
{
    public class LTurtle
    {
        public Vertex Position;
        public Vertex Orientation;      // HLU (see TABOP p.19) (Heading, Left, Up)
        public Matrix PendingRotation;
        public float SegmentDiameter;
        public int ColourIndex;

        public enum Operation : byte
        {
            MoveForwardAndDraw          = (byte) 'F',
            MoveForwardNotDraw          = (byte) 'f',
            TurnLeft                    = (byte) '+',
            TurnRight                   = (byte) '-',
            PitchUp                     = (byte) '^',
            PitchDown                   = (byte) '&',
            RollLeft                    = (byte) '\\',
            RollRight                   = (byte) '/',
            TurnAround                  = (byte) '|',
            RotateToVertical            = (byte) '$',
            BranchStart                 = (byte) '[',
            BranchEnd                   = (byte) ']',
            PolygonStart                = (byte) '{',   // Assume GL_TRIANGLE_FAN. Might need to set faces to CW.
            MoveForwardAndDraw_NoVertex = (byte) 'G',
            RecordVertex                = (byte) '.',
            PolygonEnd                  = (byte) '}',
            PredefinedSurface           = (byte) '~',
            SegmentDiameterDecrement    = (byte) '!',
            ColourIndexeIncrement       = (byte) '\'',
            CutBranch                   = (byte) '%'
        }

        public LTurtle()
        {
            Position = new Vertex(0, 0, 0);
            Orientation = new Vertex(0, 1, 0);
            PendingRotation = new Matrix(Matrix.Identity(3));
            SegmentDiameter = 3.0f;
            ColourIndex = 0;
        }
        public LTurtle(LTurtle t)
        {
            Position = new Vertex(t.Position.X, t.Position.Y, t.Position.Z);
            Orientation = new Vertex(t.Orientation.X, t.Orientation.Y, t.Orientation.Z);
            PendingRotation = new Matrix(t.PendingRotation);
            SegmentDiameter = t.SegmentDiameter;
            ColourIndex = t.ColourIndex;
        }

        public void Do(Operation op)
        {
            switch (op)
            {
                case Operation.MoveForwardAndDraw :
                    MoveForward(true);
                    break;
                case Operation.MoveForwardNotDraw :
                    MoveForward(false);
                    break;

                case Operation.TurnLeft :
                    Turn( LSystem.I().RotationAngle, 'U');
                    break;
                case Operation.TurnRight :
                    Turn(-LSystem.I().RotationAngle, 'U');
                    break;
                case Operation.TurnAround:
                    Turn(180, 'U');
                    break;
                case Operation.PitchUp:
                    Turn(-LSystem.I().RotationAngle, 'L');
                    break;
                case Operation.PitchDown:
                    Turn( LSystem.I().RotationAngle, 'L');
                    break;
                case Operation.RollLeft:
                    Turn( LSystem.I().RotationAngle, 'H');
                    break;
                case Operation.RollRight:
                    Turn(-LSystem.I().RotationAngle, 'H');
                    break;
                case Operation.RotateToVertical:
                    Orientation = new Vertex(0, 1, 0);
                    break;
               
                case Operation.SegmentDiameterDecrement :
                    SegmentDiameter *= (1.0f - LSystem.I().SegmentWidthDecrement);
                    break;
                case Operation.ColourIndexeIncrement :
                    ColourIndex = (ColourIndex+1)%LSystem.I().ColourPalette.Count;
                    break;

                // TODO : 
                case Operation.PolygonStart:
                case Operation.MoveForwardAndDraw_NoVertex:
                case Operation.RecordVertex:
                case Operation.PolygonEnd:
                case Operation.PredefinedSurface:
                    break;

                // Branching handled by LSystem
                case Operation.BranchStart: break;
                case Operation.BranchEnd: break;
                case Operation.CutBranch: break;

                default: 
                    break;
            }
        }

        // TODO: Support for parametric _operations_
        public void DoParam(Operation op, float p) { Do(op); }

        protected virtual void MoveForward(Boolean isDrawing)
        {
            var Geom = LSystem.I().Geometry;

            if (LSystem.I().RingAdjustment != null)
                Turn(LSystem.I().RingAdjustment.Value, 'H');

            var InitialOrientation = Orientation;
            if (LSystem.I().CurveAdjustOnForward)
            {
                Orientation = Orientation * PendingRotation;
                Orientation.UnitLength();

                PendingRotation = new Matrix(Matrix.Identity(3));
            }

            var InitialPosition = Position;
            Position += Orientation * LSystem.I().StepSize;
            if (!isDrawing) 
                return;

            var line = new DrawObject.DO_Line();
            line.Width = SegmentDiameter;
            line.P1 = new Vertex(InitialPosition);
            line.P2 = new Vertex(Position);
            if (LSystem.I().ColourPalette.Count > ColourIndex)
                line.Colour = LSystem.I().ColourPalette[ColourIndex];
            // TODO: Spline or something.
            //if (!Orientation.Equals(InitialOrientation))
            //    line.

            Geom.Add(line);
        }

        // Initial rotation operation attempt.
        // The other rotation functions used below are more inline with TABOP.
        private Vertex NormalizedMatrixRotationAboutOrigin(Vertex point, Vertex axis, double angle)
        {
            // See S5.2 : <http://inside.mines.edu/fs_home/gmurray/ArbitraryAxisRotation/>
            Vertex result = new Vertex();
            double common = ( axis.X * point.X + axis.Y * point.Y + axis.Z * point.Z ) * (1 - Math.Cos(angle));
            result.X = (float)(   (axis.X * common) + (point.X * Math.Cos(angle)) + ( ((-axis.Z * point.Y) + (axis.Y * point.Z)) * Math.Sin(angle) )   );
            result.Y = (float)(   (axis.Y * common) + (point.Y * Math.Cos(angle)) + ( (( axis.Z * point.X) + (axis.X * point.Z)) * Math.Sin(angle) )   );
            result.Z = (float)(   (axis.Z * common) + (point.Z * Math.Cos(angle)) + ( ((-axis.Y * point.X) + (axis.X * point.Y)) * Math.Sin(angle) )   );
            return result;
            // Call:  //Orientation = NormalizedMatrixRotationAboutOrigin(Orientation, Upwards, rotation * (Math.PI/180));
        }
        
        protected void Turn(float rotation, char turnDir)
        {
            float r = (float)(rotation * (Math.PI / 180));
            Matrix Rot_M = new Matrix(3,3);

            switch (turnDir)
            {
                case 'H':
                    Rot_M = new Matrix( new double[,]
                    {
                        {1, 0,                  0},
                        {0, (float)Math.Cos(r), -(float)Math.Sin(r)},
                        {0, (float)Math.Sin(r), (float)Math.Cos(r)}
                    });
                    break;

                case 'L':
                    Rot_M = new Matrix( new double[,]
                    {
                        {(float)Math.Cos(r),    0,  (float)-Math.Sin(r)},
                        {0,                     1,  0},
                        {(float)Math.Sin(r),    0,  (float)Math.Cos(r)}
                    });
                    break;

                case 'U':
                    Rot_M = new Matrix( new double[,]
                    {
                        {(float)Math.Cos(r),    (float)Math.Sin(r), 0},
                        {(float)-Math.Sin(r),   (float)Math.Cos(r), 0},
                        {0,                     0,                  1}
                    });
                    break;
            }

            if(LSystem.I().CurveAdjustOnForward)
            {
                PendingRotation = PendingRotation * Rot_M;
            }
            else
            {
                Orientation = Orientation * Rot_M;
                Orientation.UnitLength();
            }
        }

      
        //...

        
    }
}
