﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SharpGL.SceneGraph;

namespace LindenmayerTool
{
    // Extends the LTurtle class with methods and members relating to audio events
    public class LTurtle_Musical : LTurtle
    {
        // The Audio event tied to this turtle state
        public SoundObject Sound;
        // Child turtle; allows traversal of children with branches
        public List<LTurtle_Musical> ChildBranches = new List<LTurtle_Musical>();

        // The direction of pitch increments. 
        // Direction can be swapped (maybe for arpeggios/scales) with commands.
        Boolean isPitchAdjustPositive = true;


        public LTurtle_Musical() 
        {
            Sound = new SoundObject();
        }
        public LTurtle_Musical(LTurtle_Musical t) : base(t) 
        {
            Sound = new SoundObject(t.Sound);
        }

        // Custom mapping of characters to audio related adjustment commands
        public void DoSound(Operation op)
        {
            var ls = (LSystem_Musical)(LSystem.I());
            switch (op)
            {
                // Note Events (timing)
                case Operation.MoveForwardAndDraw:
                    AdvanceTime(true);
                    break;
                case Operation.MoveForwardNotDraw:
                    AdvanceTime(false);
                    break;

                // Sound Pitch Adjustment
                case Operation.TurnLeft:
                    Sound.Pitch = MakeAdjustment(Sound.Pitch, ls.PitchAdjustment,  this.isPitchAdjustPositive, ls.PitchAdjustmentType);
                    if (Sound.Pitch < 1) Sound.Pitch = 1;
                    break;
                case Operation.TurnRight:
                    Sound.Pitch = MakeAdjustment(Sound.Pitch, ls.PitchAdjustment, !this.isPitchAdjustPositive, ls.PitchAdjustmentType);
                    if (Sound.Pitch < 1) Sound.Pitch = 1;
                    break;
                case Operation.TurnAround:
                    isPitchAdjustPositive = !isPitchAdjustPositive;
                    break;
                case Operation.RotateToVertical:
                    isPitchAdjustPositive = true;
                    break;

                // Duration
                case Operation.RollLeft:
                    Sound.Duration = MakeAdjustment(Sound.Duration, ls.DurationAdjustment,  true, ls.DurationAdjustmentType);
                    break;
                case Operation.RollRight:
                    Sound.Duration = MakeAdjustment(Sound.Duration, ls.DurationAdjustment, false, ls.DurationAdjustmentType);
                    break;

                // Custom Parameter Adjust
                case Operation.PitchUp:
                    Sound.ParamC1 = MakeAdjustment(Sound.ParamC1, ls.ParamC1Adjustment, true, ls.ParamC1AdjustmentType);
                    break;
                case Operation.PitchDown:
                    Sound.ParamC1 = MakeAdjustment(Sound.ParamC1, ls.ParamC1Adjustment, false, ls.ParamC1AdjustmentType);
                    break;
                case Operation.PolygonStart:
                    Sound.ParamC2 = MakeAdjustment(Sound.ParamC2, ls.ParamC2Adjustment, true, ls.ParamC2AdjustmentType);
                    break;
                case Operation.PolygonEnd:
                    Sound.ParamC2 = MakeAdjustment(Sound.ParamC2, ls.ParamC2Adjustment, false, ls.ParamC2AdjustmentType);
                    break;
                case Operation.PredefinedSurface:
                    Sound.ParamC3 = MakeAdjustment(Sound.ParamC3, ls.ParamC3Adjustment, true, ls.ParamC3AdjustmentType);
                    break;
                case Operation.RecordVertex:
                    Sound.ParamC3 = MakeAdjustment(Sound.ParamC3, ls.ParamC3Adjustment, false, ls.ParamC3AdjustmentType);
                    break;

                // Volume
                case Operation.SegmentDiameterDecrement:
                    Sound.Volume = MakeAdjustment(Sound.Volume, ls.VolumeAdjustment, false, ls.VolumeAdjustmentType);
                    break;

                // Instrument
                case Operation.ColourIndexeIncrement:
                    Sound.Instrument = (Sound.Instrument + 1) % LSystem.I().ColourPalette.Count;
                    break;

                // Branching handled by LSystem
                case Operation.BranchStart: break;
                case Operation.BranchEnd: break;
                case Operation.CutBranch: break;

                default:
                    break;
            }
        }

        // Returns an adjusted value based on the current value, change amount, 
        // change direction, and adjustment behaviour.
        public static float MakeAdjustment(float initial, float delta, bool isPositive, LSystem_Musical.AdjustmentType type)
        {
            switch (type)
            {
                case LSystem_Musical.AdjustmentType.Additive:
                    return initial + (isPositive ? 1 : -1) * delta;
                    
                case LSystem_Musical.AdjustmentType.Multiplicative:
                    return initial * (1.0f + ((isPositive ? 1 : -1) * delta));

                default:
                    return initial;
            }
        }

        // Scales and adjusts a coordinate value for use as an audio paramater
        public static float ValueFromAxis(float axisval, float init, float delta, LSystem_Musical.AdjustmentType type)
        {
            float ret = init;

            if( ((LSystem_Musical)(LSystem.I())).isAxisValuesAbsolute )
                axisval = (axisval < 0) ? -axisval : axisval;

            switch (type)
            {
                case LSystem_Musical.AdjustmentType.Additive:
                    ret += axisval;
                    break;

                case LSystem_Musical.AdjustmentType.Multiplicative:
                    ret += axisval * delta;
                    break;

                default:
                    return axisval;
            }

            return ret;
        }

        // Overrode MoveForward to allow embedding timing data into DrawObjects
        protected override void MoveForward(Boolean isDrawing)
        {
            base.MoveForward(isDrawing);
            if (!isDrawing) return;

            var Geom = LSystem.I().Geometry;
            var drawObj = Geom[Geom.Count-1];
            drawObj.DrawTime = (int)Sound.StartTime;
        }

        // Updates the sound-related parameters of the turtle and adds a new sound event.
        public void AdvanceTime(bool isMakingSound)
        {
            var ls = (LSystem_Musical)(LSystem.I());

            if(isMakingSound && ls.NoteStepCreateMetric == LSystem_Musical.NoteStepCreationMetric.StartStep)
                ls.Score.Add(Sound);

            ManageTimestamp();
            UpdateSoundParams();

            if (isMakingSound && ls.NoteStepCreateMetric == LSystem_Musical.NoteStepCreationMetric.EndStep)
                ls.Score.Add(Sound);
        }

        // Determine the starting time of any events created with the current turtle
        private float ManageTimestamp()
        {
            LSystem_Musical ls = (LSystem_Musical)(LSystem.I());
            
            switch(ls.TimingMetric)
            {
                
                case LSystem_Musical.TimeMetric.TotalStepLength:
                    Sound.StartTime += (float)(Math.Round(ls.StepSize * ls.StepTimeScale));
                    break;
                case LSystem_Musical.TimeMetric.Coord_X:
                    Sound.StartTime = Position.X * ls.StepTimeScale;
                    break;
                case LSystem_Musical.TimeMetric.Coord_Y:
                    Sound.StartTime = Position.Y * ls.StepTimeScale;
                    break;
                case LSystem_Musical.TimeMetric.Coord_Z:
                    Sound.StartTime = Position.Z * ls.StepTimeScale;
                    break;
                case LSystem_Musical.TimeMetric.OriginDistance3D:
                    Sound.StartTime = (float)Position.Magnitude() * ls.StepTimeScale;
                    break;
                case LSystem_Musical.TimeMetric.OriginDistance2D:
                    Vertex v = new Vertex(Position.X, Position.Y, 0);
                    Sound.StartTime = (float)v.Magnitude() * ls.StepTimeScale;
                    break;
                default: break;
            }

            return Sound.StartTime;
        }

        // Get updated sound source values if mapped to position coordinates
        private void UpdateSoundParams()
        {
            var ls = (LSystem_Musical)(LSystem.I());

            switch (ls.PitchDeterminationType)
            {
                case LSystem_Musical.IncrementOrAxis.Coord_X: 
                    Sound.Pitch = ValueFromAxis(Position.X, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Y:
                    Sound.Pitch = ValueFromAxis(Position.Y, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Z:
                    Sound.Pitch = ValueFromAxis(Position.Z, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                default: break;
            }
            switch (ls.DurationDeterminationType)
            {
                case LSystem_Musical.IncrementOrAxis.Coord_X:
                    Sound.Duration = ValueFromAxis(Position.X, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Y:
                    Sound.Duration = ValueFromAxis(Position.Y, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Z:
                    Sound.Duration = ValueFromAxis(Position.Z, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                default: break;
            }
            switch (ls.VolumeDeterminationType)
            {
                case LSystem_Musical.IncrementOrAxis.Coord_X:
                    Sound.Volume = ValueFromAxis(Position.X, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Y:
                    Sound.Volume = ValueFromAxis(Position.Y, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Z:
                    Sound.Volume = ValueFromAxis(Position.Z, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                default: break;
            }

            switch (ls.ParamC1DeterminationType)
            {
                case LSystem_Musical.IncrementOrAxis.Coord_X:
                    Sound.ParamC1 = ValueFromAxis(Position.X, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Y:
                    Sound.ParamC1 = ValueFromAxis(Position.Y, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Z:
                    Sound.ParamC1 = ValueFromAxis(Position.Z, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                default: break;
            }
            switch (ls.ParamC2DeterminationType)
            {
                case LSystem_Musical.IncrementOrAxis.Coord_X:
                    Sound.ParamC2 = ValueFromAxis(Position.X, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Y:
                    Sound.ParamC2 = ValueFromAxis(Position.Y, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Z:
                    Sound.ParamC2 = ValueFromAxis(Position.Z, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                default: break;
            }
            switch (ls.ParamC3DeterminationType)
            {
                case LSystem_Musical.IncrementOrAxis.Coord_X:
                    Sound.ParamC3 = ValueFromAxis(Position.X, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Y:
                    Sound.ParamC3 = ValueFromAxis(Position.Y, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                case LSystem_Musical.IncrementOrAxis.Coord_Z:
                    Sound.ParamC3 = ValueFromAxis(Position.Z, ls.PitchDefault, ls.PitchAdjustment, ls.PitchAdjustmentType);
                    break;
                default: break;
            }



        }

    }
}
