﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using SharpGL;
using SharpGL.SceneGraph;
using System.IO;
using System.Diagnostics;
using System.Media;

namespace LindenmayerTool
{
    // Interaction logic for MainWindow.xaml
    public partial class MainWindow : Window
    {
        private double Aspect = 1.0f;
        private Vertex ViewTranslation = new Vertex(0.0f, 0.0f, 0.0f);


        private static MainWindow _main = null;
        public static MainWindow I()
        {
            if (_main == null) _main = new MainWindow();
            return _main;
        }
        
        public MainWindow()
        {
            _main = this;
            InitializeComponent();
            InitDefaultSystem();
        }

        public void InitDefaultSystem()
        {
            var lSys = LSystem.I();
            var MusicOpts = MusicMenu.I();

            lSys.Seed = "X";
            lSys.Rules.Add(new ProductionRule("X", "F-[[!X]+!X]+F[+F!X]-!X", 1));
            lSys.Rules.Add(new ProductionRule("F", "FF", 1));
            lSys.Iterations = 6;
            lSys.StepSize = 3;
            lSys.RotationAngle = 22.5f;
            lSys.SegmentWidthInitial = 4.0f;
            lSys.SegmentWidthDecrement = 0.4f;
            lSys.DoAll();

            this.ViewTranslation = new Vertex(-24, 227, 120);
        }

        private void Window_OnSizeChanged(object sender, SizeChangedEventArgs e)
        {
            GLCanvas.Width = GLContainer.Width;
            GLCanvas.Height = GLContainer.Height;
            Aspect = GLCanvas.Width / GLCanvas.Height;
            if (Double.IsNaN(Aspect)) Aspect = 1.0f;
        }

        private void Window_OnClosed(object sender, EventArgs e)
        {
            System.Environment.Exit(0);
        }

        private void OpenGLControl_OpenGLDraw(object sender, OpenGLEventArgs args)
        {
            OpenGL gl = args.OpenGL;
            gl.Clear(OpenGL.GL_COLOR_BUFFER_BIT | OpenGL.GL_DEPTH_BUFFER_BIT);
            gl.LoadIdentity();
            gl.Ortho(-200, 200, -(200*Aspect), (200*Aspect), 1,-100);
            gl.Translate(-ViewTranslation.X, -ViewTranslation.Y, -ViewTranslation.Z);

            var lSys = LSystem.I();
            if (!timer.IsRunning)
                lSys.Render(gl);
            else
                ((LSystem_Musical)lSys).Render(gl,timer.ElapsedMilliseconds);

            gl.Flush();

            // Apparently SharpGL likes to eat memory and crash 
            // (1680x1050 Gets to 1.5Gb in ~2s).
            // Now down to a stable ~70Mb
            GC.Collect();
        }

        Nullable<Vertex> OldViewTranslation = null;
        Vector OldMouseCoords;
        private void OpenGLControl_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.LeftButton == MouseButtonState.Pressed)
            {
                double CurX = e.GetPosition(GLCanvas).X;
                double CurY = e.GetPosition(GLCanvas).Y;
                if (OldViewTranslation == null){
                    OldViewTranslation = new Vertex(ViewTranslation.X, ViewTranslation.Y, ViewTranslation.Z);
                    OldMouseCoords = new Vector(CurX, CurY);
                }
                Vector MouseDragDelta = new Vector(OldMouseCoords.X - CurX, OldMouseCoords.Y - CurY);

                ViewTranslation.X = (float)( OldViewTranslation.Value.X + MouseDragDelta.X );
                ViewTranslation.Y = (float)( OldViewTranslation.Value.Y - MouseDragDelta.Y );
            }
            else if (OldViewTranslation != null)
                OldViewTranslation = null;
        }
        private void OpenGLControl_MouseScroll(object sender, MouseWheelEventArgs e)
        {
            ViewTranslation.Z += e.Delta / 10.0f;
        }
        private void OpenGLControl_MouseDown(object sender, MouseButtonEventArgs e)
        {
            if (e.MiddleButton == MouseButtonState.Pressed)
                ViewTranslation = new Vertex(0, 0, 0);
        }

        private void Menu_Click_Open(object sender, RoutedEventArgs e)
        {
            Microsoft.Win32.OpenFileDialog fileDialog = new Microsoft.Win32.OpenFileDialog() 
                {   
                    Filter = "Lindenmayer Systems (*.lsys)|*.lsys|All Files (*.*)|*.*", 
                    InitialDirectory = System.Environment.CurrentDirectory
                };
            bool? res = fileDialog.ShowDialog();
            if (res == null || res.Value==false ) return;
            string filename = fileDialog.FileName;
            var lSys = LSystem.LoadFromFile(filename);
            LSystem.SetActive(lSys);
            lSys.DoAll();
        }

        private void Menu_Click_Refresh(object sender, RoutedEventArgs e)
        {
            var lSys = LSystem.I();
            lSys.DoAll();
        }

        private void Menu_Click_RNG(object sender, RoutedEventArgs e)
        {
            SeedPrompt.I().Show();
        }

        private void Menu_Click_ExportObj(object sender, RoutedEventArgs e)
        {
            Microsoft.Win32.SaveFileDialog fileDialog = new Microsoft.Win32.SaveFileDialog()
            {
                Filter = "Wavefront Object (*.obj)|*.obj|All Files (*.*)|*.*",
                InitialDirectory = System.Environment.CurrentDirectory,
                OverwritePrompt = true
            };
            bool? res = fileDialog.ShowDialog();
            if (res == null || res.Value == false) return;
            string filename = fileDialog.FileName;
            var Exporter = new ObjExporter();
            Exporter.Export(LSystem.I(), filename);
        }

        private void Menu_Click_ExportSco(object sender, RoutedEventArgs e)
        {
            Microsoft.Win32.SaveFileDialog fileDialog = new Microsoft.Win32.SaveFileDialog()
            {
                Filter = "Csound Score (*.sco)|*.sco|All Files (*.*)|*.*",
                InitialDirectory = System.Environment.CurrentDirectory,
                OverwritePrompt = true
            };
            bool? res = fileDialog.ShowDialog();
            if (res == null || res.Value == false) return;
            string filename = fileDialog.FileName;
            var Exporter = new ScoExporter();
            Exporter.Export(LSystem.I(), filename);
        }

        private void Menu_Click_MusicSettings(object sender, RoutedEventArgs e)
        {
            MusicMenu.I().Show();
        }

        Stopwatch timer = new Stopwatch();
        SoundPlayer player = new SoundPlayer();
        public void Menu_Click_Playback_Start(object sender, RoutedEventArgs e)
        {
            player.Stop();
            player.Dispose();

            CreateWavefile();

            String tmp = GetTempPath();
            String wavFile = Path.Combine( tmp, "out.wav" );

            if (!File.Exists(wavFile)) return;

            player = new SoundPlayer(wavFile);
            player.Load();

            timer.Restart();
            player.Play();
        }
        public void Menu_Click_Playback_Stop(object sender, RoutedEventArgs e)
        {
            player.Stop();
            timer.Stop();
        }


        public static Boolean CreateWavefile()
        {
            var scoPath = Path.Combine(GetTempPath(), "out.sco");
            var outPath = Path.Combine(GetTempPath(), "out.wav");

            // Export the score
            var Exporter = new ScoExporter();
            Exporter.Export(LSystem.I(), scoPath );

            // Prepare to invoke CSound with the given flags and files
            var flags = "-d -W"; // non-verbose, WAV output
            var ps = new ProcessStartInfo(MusicMenu.CS_Bin_File, flags+" "+QuoteWrap(MusicMenu.CS_Orc_File)+" "+QuoteWrap(scoPath)+" -o "+QuoteWrap(outPath) );
            
            try
            {
                var p = Process.Start(ps);
                p.WaitForExit(); // Do synchronously. Don't play before the file's created.
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
                return false;
            }

            return true;
        }

        public static String QuoteWrap(String ins)
        {
            return "\"" + ins + "\"";
        }
        public static String GetTempPath()
        {
            return Path.Combine(System.IO.Path.GetTempPath(), "LindenmayerTool");
        }

        private void Menu_Click_AxisWrap(object sender, RoutedEventArgs e)
        {
            WrapOptions.I().Show();
        }
        
        
    }
}
