﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SharpGL.SceneGraph;
using System.IO;

namespace LindenmayerTool
{
    public class ObjExporter
    {
        // Settings
        public bool ConsolidateObjects = true;
        public bool AddSegmentEndFaces = true;
        public bool AddSegmentMergeFaces = false;
        public int SegmentFaces = 3;

        private int VerticeCount;   // Current vertice number in processing stage

        // Vertex equality tester making use of custom epsilon.
        static bool VertexEqual(Vertex v1, Vertex v2)
        {
            const float e = 0.00000001f; //Epsilon
            if (Math.Abs(v1.X - v2.X) > e) return false;
            if (Math.Abs(v1.Y - v2.Y) > e) return false;
            if (Math.Abs(v1.Z - v2.Z) > e) return false;
            return true;
        }

        // Merges connected line segments of the same width.
        // Reduces number of vertices and faces in exported object.
        List<DrawObject> ConsolidateDrawObjects(List<DrawObject> objsList) 
        {
            var g = new List<DrawObject>();
            var objs = new List<DrawObject>();
            objsList.ForEach(o => objs.Add(o));

            // For each line
            for (int i = 0; i < objs.Count; i++)
            {
                var dObj = objs[i];
                if (dObj.GetType() != typeof(DrawObject.DO_Line)) continue; // do other stuff...
                DrawObject.DO_Line l = (DrawObject.DO_Line) dObj;

                // Get points and directions
                Vertex P1 = new Vertex(l.P1.X, l.P1.Y, l.P1.Z);
                Vertex P2 = new Vertex(l.P2.X, l.P2.Y, l.P2.Z);
                Vertex d1 = new Vertex(P2.X - P1.X, P2.Y - P1.Y, P2.Z - P1.Z);
                Vertex d2 = new Vertex(P1.X - P2.X, P1.Y - P2.Y, P1.Z - P2.Z);
                d1.UnitLength();
                d2.UnitLength();

                bool merged = true;
                // While changes were made (or we haven't checked yet) try to merge lines...
                while (merged)
                {
                    merged = false;
                    for (int j = i + 1; j < objs.Count; j++)
                    {
                        var dObj2 = objs[j];
                        if (dObj2.GetType() != typeof(DrawObject.DO_Line)) continue; // do other stuff...
                        DrawObject.DO_Line l2 = (DrawObject.DO_Line)dObj2;
                   
                        Vertex Q1 = new Vertex(l2.P1.X, l2.P1.Y, l2.P1.Z);
                        Vertex Q2 = new Vertex(l2.P2.X, l2.P2.Y, l2.P2.Z);
                        Vertex d = new Vertex(Q2.X - Q1.X, Q2.Y - Q1.Y, Q2.Z - Q1.Z);
                        d.UnitLength();

                        // If the lines extend in the correct direction, merge with them
                        if (!l.Width.Equals(l2.Width)) continue;
                        if ( VertexEqual(d,d1) ) {
                            if ( VertexEqual(P2,Q1) ) {
                                P2 = Q2;
                                merged = true;
                            }
                            if ( VertexEqual(P1,Q2) ) {
                                P1 = Q1;
                                merged = true;
                            }
                        }
                        if ( VertexEqual(d,d2) ) {
                            if ( VertexEqual(P2,Q2) ) {
                                P2 = Q1;
                                merged = true;
                            }
                            if (VertexEqual(P1, Q1))
                            {
                                P1 = Q2;
                                merged = true;
                            }
                        }
                        if (merged)
                        {
                            objs.RemoveAt(j);
                            break;
                        }
                    }//dObj2
                }//merges?

                var newL = new DrawObject.DO_Line();
                newL.Width = l.Width;
                newL.P1 = P1;
                newL.P2 = P2;
                g.Add(newL);
            }
            return g;
        }

        // Formats a vertex definition for obj
        String Def_Vertex(Vertex v)
        {
            return "v " + v.X + " " + v.Y + " " + v.Z + Environment.NewLine;
        }
        // Formats a face definition for obj
        String Def_Face(int[] vs)
        {
            String fd = "f ";
            for (int i = 0; i < vs.Length; i++)
                fd += (vs[i]).ToString() + " ";
            return fd + Environment.NewLine;
        }

        Vertex RotationOffsetAboutAxis(Vertex axis, float extension, float angle)
        {
            // See S5.2 : <http://inside.mines.edu/fs_home/gmurray/ArbitraryAxisRotation/>
            Vertex point = new Vertex(extension, 0, 0);
            Vertex result = new Vertex();
            double common = (axis.X * point.X + axis.Y * point.Y + axis.Z * point.Z) * (1 - Math.Cos(angle));
            result.X = (float)((axis.X * common) + (point.X * Math.Cos(angle)) + (((-axis.Z * point.Y) + (axis.Y * point.Z)) * Math.Sin(angle)));
            result.Y = (float)((axis.Y * common) + (point.Y * Math.Cos(angle)) + (((axis.Z * point.X) + (axis.X * point.Z)) * Math.Sin(angle)));
            result.Z = (float)((axis.Z * common) + (point.Z * Math.Cos(angle)) + (((-axis.Y * point.X) + (axis.X * point.Y)) * Math.Sin(angle)));
            return result;
        }

        String GenerateSegment(DrawObject.DO_Line l)
        {
            StringBuilder sb = new StringBuilder();

            // Direction unit vector
            Vertex dir = new Vertex(Math.Abs(l.P2.X - l.P1.X), Math.Abs(l.P2.Y - l.P1.Y), Math.Abs(l.P2.Z - l.P1.Z)); 
            dir.UnitLength();
            
            // Number of vertices added for this object
            int vc = 4; // vertice count
            if (SegmentFaces > 2) vc += (2 * (SegmentFaces - 2));
            int vh = vc / 2; // half vertice count
            float rot_step = (float)((360.0f / SegmentFaces)* (Math.PI / 180));
            Vertex[] v = new Vertex[vc];
            
            //    5
            //  /  \
            // /    \
            // 4-----3
            //  \     \2        OBJ requires
            //   \   / \        CCW vertex order
            //    \ /   \
            //     1-----0

            for( int i = 0; i < vh; i++ )
            {
                Vertex off = RotationOffsetAboutAxis(dir, l.Width/2, i*rot_step);
                v[i]             = l.P1 + off;
                v[i+(vh)] = l.P2 + off;
            }
            
            // Vertices:
            for (int i = 0; i < vc; i++)
                sb.Append(Def_Vertex(v[i]));

            // Faces:
            int vfirst = VerticeCount;
            int vcur = vfirst;
            int[] vs = new int[4];
            for (int i = 0; i < (vh-1); i++)
            {  
                vs[0] = vcur+0;
                vs[1] = vcur+0+vh;
                vs[2] = vcur+1+vh;
                vs[3] = vcur+1;
                sb.Append(Def_Face(vs));
                vcur += 1;
            }
            vs[0] = vcur+0;
            vs[1] = vcur+0+vh;
            vs[2] = vfirst+0+vh;
            vs[3] = vfirst+0;            
            sb.Append(Def_Face(vs));
            if (AddSegmentEndFaces)
            {
                vs = new int[vh];
                for (int i = 0; i < vh; i++)
                    vs[i] = vfirst+i ;
                sb.Append(Def_Face(vs));
                vs = new int[vh];
                for (int i = vh; i < vc; i++)
                    vs[i-vh] = vfirst+i+vh;
                sb.Append(Def_Face(vs));
            }

            VerticeCount += vc;
            return sb.ToString();
        }

        String GenerateObjFile(LSystem lsys)
        {
            List<DrawObject> Geometry = lsys.Geometry;
            if (ConsolidateObjects)
                Geometry = ConsolidateDrawObjects(lsys.Geometry);
            
            StringBuilder sb = new StringBuilder();
            VerticeCount = 1;

            foreach (var dObj in Geometry)
            {
                if (dObj.GetType() != typeof(DrawObject.DO_Line)) continue; // do other stuff...
                DrawObject.DO_Line l = (DrawObject.DO_Line) dObj;
                sb.Append(GenerateSegment(l));
            }

            return sb.ToString();
        }

        public void Export(LSystem lsys, String filename)
        {
            var objString = GenerateObjFile(lsys);
            File.WriteAllText(filename, objString);
        }

    }
}
